from typing import Generic, TypeVar

from rainbowadn.collection.collectioninterface import *
from rainbowadn.core import *
from rainbowadn.nullability import *

__all__ = ('BlockCollectionInterface',)

BlockType = TypeVar('BlockType')
HeaderType = TypeVar('HeaderType')


class BlockCollectionInterface(
    CollectionInterface[
        BlockType
    ],
    Generic[BlockType, HeaderType]
):
    def __init__(
            self,
            reference: NullableReference[BlockType]
    ):
        assert isinstance(reference, NullableReference)
        super().__init__(reference)

    async def _verify_link(
            self,
            block: BlockType,
            previous: NullableReference[BlockType]
    ) -> bool:
        raise NotImplementedError

    async def verify_link(
            self,
            previous: NullableReference[BlockType]
    ) -> bool:
        assert isinstance(previous, NullableReference)
        if self.reference.null():
            return True
        else:
            return await self._verify_link(
                await self.reference.resolve(),
                previous
            )

    async def add(
            self,
            header: HashPoint[HeaderType]
    ) -> 'BlockCollectionInterface[BlockType, HeaderType]':
        raise NotImplementedError
