from typing import TypeVar

from rainbowadn.chain.blockchainprotocol import *
from rainbowadn.chain.derivation import *
from rainbowadn.chain.stages import *
from rainbowadn.core import *
from .reducible import *
from .reductionprotocol import *
from .reductionstageprotocol import *

__all__ = ('ReductionChainProtocol',)

ReductorType = TypeVar('ReductorType')
AccumulatorType = TypeVar('AccumulatorType')


class ReductionChainProtocol(
    BlockChainProtocol[
        ReductorType,
        StateStage[ReductorType, AccumulatorType, Reducible[ReductorType, AccumulatorType]]
    ]
):
    def __init__(
            self,
            protocol: ReductionProtocol[ReductorType, AccumulatorType],
            reductor_factory: RainbowFactory[ReductorType],
            accumulator_factory: RainbowFactory[AccumulatorType],
    ):
        assert isinstance(protocol, ReductionProtocol)
        assert isinstance(reductor_factory, RainbowFactory)
        assert isinstance(accumulator_factory, RainbowFactory)

        reducible_factory: RainbowFactory[
            Reducible[ReductorType, AccumulatorType]
        ] = ReducibleFactory(
            reductor_factory, accumulator_factory
        )
        assert isinstance(reducible_factory, RainbowFactory)
        stage_protocol: ActiveStageProtocol[
            ReductorType,
            AccumulatorType,
            Reducible[ReductorType, AccumulatorType]
        ] = ReductionStageProtocol(protocol)
        assert isinstance(stage_protocol, ActiveStageProtocol)

        super().__init__(
            ActiveStageStateProtocol(
                stage_protocol,
                reducible_factory,
                accumulator_factory,
            ),
            reductor_factory,
            StateStageFactory(
                stage_protocol,
                reducible_factory,
                accumulator_factory
            )
        )

        self.reductor_factory = reductor_factory
        self.accumulator_factory = accumulator_factory
