from typing import Any, Generic, TypeAlias, TypeVar

from rainbowadn.atomic import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.collection.trees.binary.actions import *
from rainbowadn.collection.trees.binary.core import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from ._binaryflow import *
from ._flowtree import *

__all__ = ('FlowStandard',)

KeyT = TypeVar('KeyT')
FS: TypeAlias = 'FlowStandard[KeyT]'
ABT: TypeAlias = 'ActiveBinaryTree[KeyT, Integer]'
BP: TypeAlias = 'BinaryProtocolized[KeyT, Integer, ABT]'


class FlowStandard(
    FlowTree[KeyT, FS],
    Generic[KeyT]
):
    def __init__(self, protocolized: BP):
        assert isinstance(protocolized, BinaryProtocolized)
        self.protocolized = protocolized

    async def contains(self, key: KeyT) -> bool:
        return await ContainsAction(key).on(self.protocolized)

    def _protocolized(self: FS) -> BP:
        return self.protocolized

    @classmethod
    def _protocolized_mapper(cls) -> Mapper[FS, BP]:
        return CallableMapper(cls._protocolized)

    async def verify_subset(self, trees: Reducer[FS, CheckResult]) -> bool:
        assert isinstance(trees, Reducer)
        reducer: Reducer[BP, CheckResult] = MapReducer(self._protocolized_mapper(), trees)
        assert_true(await VerifySubsetAction(reducer).on(self.protocolized))
        return True

    async def reducer(self) -> Reducer[KeyT, Any]:
        return BinaryReducer(self.protocolized).loose()
