import unittest

import nacl.signing

from rainbowadn.chain import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from rainbowadn.testing.resolvers import *
from rainbowadn.v13 import *


class TestV13(unittest.IsolatedAsyncioTestCase):
    async def test_bankchain(self):
        set_gather_linear()
        with self.subTest('create empty'):
            bank: BankChain = BankChain.empty(ReductionChainMetaFactory().loose())
        with self.subTest('prepare transactions'):
            key_0 = nacl.signing.SigningKey.generate()
            transaction_0 = Transaction.make(
                [],
                [CoinData.of(Subject(key_0.verify_key), 1_000_000)],
                []
            )
            coin_0, coin_1 = await transaction_0.coins(MINT_CONST, NotNull(HashPoint.of(Subject(key_0.verify_key))))
        with self.subTest('add transactions'):
            bank = await bank.adds(
                [
                    transaction_0,
                    Transaction.make(
                        [coin_1],
                        [CoinData.of(Subject(nacl.signing.SigningKey.generate().verify_key), 10_000)],
                        [key_0]
                    ),
                ]
            )
        with self.subTest('add empty'):
            bank = await bank.adds(
                []
            )
            print(await bank.reference.str(0))
        with self.subTest('verify'):
            assert_true(await bank.verify())
        with self.subTest('recover'):
            bank = BankChain.from_reference(
                ReductionChainMetaFactory(), await default_resolver().migrate_resolved(bank.reference)
            )
            set_gather_asyncio()
            print('recovering')
            print(await bank.reference.str(0))
            print('recovered')
        with self.subTest('verify'):
            assert_true(await bank.verify())
